from __future__ import annotations

from datetime import timedelta
from typing import (
    Any,
    List,
    Optional,
    Union,
    overload,
)

import numba as nb  # type: ignore
import numpy as np
import numpy.typing as npt

__all__ = ["IndexedSnapshot"]

FloatArray = Optional[Union[npt.NDArray[np.float64], List[float]]]


class IndexedSnapshot:
    """
    This class stores timeseries in Modified Run-Length Encoding.

    Normal RLE stores the length of a run of repeated values.

    In this class, we store the index of the start of a run of repeated values.

    For example, if we have a intensity timeseries of
        [0, 0, 0, 1, 1, 0, 0, 0, 0, 0]

    then the RLE representation is
        [(3, 0), (2, 1), (5, 0)]

    and the IntensityRLE representation is
        [(0, 0), (3, 1), (5, 0)]
    """

    __slots__ = ("_time", "_intensity", "_intensity_delta")

    def __init__(
        self,
        time: FloatArray = None,
        intensity: FloatArray = None,
    ):
        """
        Parameters
        ----------
        time: Optional[Union[`np.ndarray`, `List`]]
            Time index of the intensity timeseries.

        intensity: Optional[Union[`np.ndarray`, `List`]]
            Intensity values (mm/h) of the intensity timeseries.
        """
        self._time, self._intensity = _isnapshot_check(time, intensity)
        self._intensity_delta = _intensity_delta_check(self._time, self._intensity)

    @classmethod
    def fromDelta(
        cls, time: FloatArray, intensity_delta: FloatArray
    ) -> IndexedSnapshot:
        """
        Create an IndexedSnapshot timeseries from the delta encoded intensity timeseries.

        Parameters
        ----------
        time: Optional[Union[`np.ndarray`, `List`]]
            Time index of the intensity timeseries.
            The order of the times does not matter.
            Duplicate times will be summed together. (Because it's multiple delta at the same time)

        intensity_delta: Optional[Union[`np.ndarray`, `List`]]
            Delta Encoded Intensity values (mm/h) of the intensity timeseries.

        """
        # Check if time and intensity_delta is np.ndarray
        if not isinstance(time, np.ndarray) or time.dtype != np.float64:
            time = np.array(time, dtype=np.float64)
        if (
            not isinstance(intensity_delta, np.ndarray)
            or intensity_delta.dtype != np.float64
        ):
            intensity_delta = np.array(intensity_delta, dtype=np.float64)

        time, intensity_isnapshot = _delta_to_isnapshot(time, intensity_delta)

        return cls(time, intensity_isnapshot)

    @property
    def time(self) -> npt.NDArray[np.float64]:
        """
        Return the time index of the IndexedSnapshot timeseries.

        Returns
        -------
        time : npt.NDArray[np.float64]
            1D array of time index that follows IndexedSnapshot format.
            Ending time is for indicating the end of the timeseries.
        """
        return self._time

    @property
    def intensity(self) -> npt.NDArray[np.float64]:
        """
        Return the intensity of the IndexedSnapshot timeseries.

        Returns
        -------
        intensity : npt.NDArray[np.float64]
            1D array of intensity that follows IndexedSnapshot format.
            Ending intensity is for indicating the end of the timeseries. So it's np.nan.
        """
        return self._intensity

    @property
    def intensity_delta(self) -> npt.NDArray[np.float64]:
        """
        Return the delta encoded intensity of the IndexedSnapshot timeseries.

        Returns
        -------
        intensity_delta : npt.NDArray[np.float64] with shape (n, 2)
            2D array of intensity that follows delta encoding format.
            The first column is the time index.
            The second column is the change in intensity.
        """
        return self._intensity_delta

    def add(self, timeseries: IndexedSnapshot, sequential: bool = False) -> None:
        """
        Add another IndexedSnapshot timeseries to the current IndexedSnapshot timeseries.

        Parameters
        ----------
        timeseries : IndexedSnapshot
            The IndexedSnapshot timeseries to be added.
        sequential : bool, optional
            If False, the timeseries will be added as its original time index.  So overlapping timeseries will be summed together in the overlapping region.
            If True, the timeseries will be added sequentially. So the timeseries will be shifted to the right of the current timeseries.
        """
        if len(timeseries.time) == 0:
            return
        if sequential is True:
            if len(self._time) != 0:
                # Shift the start of the timeseries to be right after th last time of the current timeseries
                timeseries = IndexedSnapshot(
                    timeseries.time - (timeseries.time[0] - self._time[-1]),
                    timeseries.intensity,
                )
        self.__iadd__(timeseries)

    def __iadd__(self, timeseries: IndexedSnapshot) -> IndexedSnapshot:
        result_isnapshot = _merge_isnapshot(self, timeseries)

        self._time, self._intensity = (
            result_isnapshot._time,
            result_isnapshot._intensity,
        )
        self._intensity_delta = result_isnapshot._intensity_delta

        return self

    def __add__(self, timeseries: IndexedSnapshot) -> IndexedSnapshot:
        return _merge_isnapshot(self, timeseries)

    @overload
    def __getitem__(self, time_idx: slice) -> IndexedSnapshot: ...

    @overload
    def __getitem__(self, time_idx: int) -> np.float64: ...

    def __getitem__(self, time_idx: Any) -> Any:
        if isinstance(time_idx, slice):
            return self._get_slice_idx(time_idx)
        elif isinstance(time_idx, int):
            return self._get_int_idx(time_idx)
        else:
            raise TypeError("time_idx must be an int or a slice")

    def _get_slice_idx(self, time_idx: slice) -> IndexedSnapshot:
        """
        ### This is an internal function. You shouldn't use it directly unless you know what you are doing.
        """
        if time_idx.start >= time_idx.stop:
            return type(self)()
        # Use binary search to find the index
        start_idx = np.searchsorted(a=self._time, v=time_idx.start, side="right")
        stop_idx = np.searchsorted(a=self._time, v=time_idx.stop, side="left")

        if start_idx == len(self._time):
            return type(self)()
        if stop_idx == 0:
            return type(self)()

        time = self._time[max(0, start_idx - 1) : stop_idx]
        intensity = self._intensity[max(0, start_idx - 1) : stop_idx]

        if start_idx != 0:
            time[0] = time_idx.start
        if stop_idx < len(self._time):
            time = np.append(time, time_idx.stop)
            intensity = np.append(intensity, np.nan)

        return type(self)(time, intensity)

    def _get_int_idx(self, time_idx: int) -> np.float64:
        """
        ### This is an internal function. You shouldn't use it directly unless you know what you are doing.
        """
        # Use binary search to find the index
        idx = np.searchsorted(a=self._time, v=time_idx, side="right")
        if idx == 0 or idx == len(self._time):
            return np.float64(0.0)
        return self._intensity[int(idx) - 1]

    def __str__(self) -> str:
        time_value = "\n".join(
            f"{self.time[i]} {self.intensity[i]}" for i in range(len(self.time))
        )
        return time_value

    def sum(self) -> float:
        """
        Calculate the total depth of the intensity timeseries.
        This value should roughly be the same after rescaling with any scale.

        Returns
        -------
        sum : float
            The sum of depth of the intensity timeseries.
        """
        if self._time.size == 0:
            return 0
        depth, _ = _isnapshot_sum_and_duration(self._time, self._intensity)
        return depth

    def mean(self) -> float:
        """
        Calculate the mean of the intensity timeseries.

        Returns
        -------
        mean : float
            The mean of the intensity timeseries.
        """
        if self._time.size == 0:
            return np.nan
        return _isnapshot_mean(self._time, self._intensity)

    def sum_squared_error(self) -> float:
        """
        Calculate the sum of squared error of the intensity timeseries.

        Returns
        -------
        sum of squared error: float
            The sum of squared error of the intensity timeseries.
        """
        if self._time.size == 0:
            return np.nan
        return _isnapshot_sum_squared_error(self._time, self._intensity)

    def autocorr_coef(self, lag: Optional[float] = 1) -> float:
        """
        Calculate the n-lag autocorrelation coefficient of the intensity timeseries.
        It's basically Cov(X_t, X_{t+lag}) / Var(X_t)

        Parameters
        ----------
        lag : int, optional
            The lag of the autocorrelation coefficient, by default 1

        Returns
        -------
        n-lag Autocorrelation coefficient: float
        """
        if self._time.size == 0:
            return np.nan
        return _isnapshot_acf(self._time, self._intensity, lag=lag)

    def variance(self, biased: bool = False) -> float:
        """
        Calculate the variance of the intensity timeseries.

        Parameters
        ----------
        biased : bool, optional
            Whether to use the biased estimator of the variance, by default False.

        Returns
        -------
        variance: float
        """
        if self._time.size == 0:
            return np.nan
        return _isnapshot_variance(self._time, self._intensity, biased=biased)

    def coef_variation(self, biased: bool = False) -> float:
        """
        Calculate the coefficient of variation of the intensity timeseries.

        Parameters
        ----------
        biased : bool, optional
            Whether to use the biased estimator of the standard deviation, by default False.

        Returns
        -------
        coefficient of variation: float
        """
        if self._time.size == 0:
            return np.nan
        return _isnapshot_coef_var(self._time, self._intensity, biased=biased)

    def skewness(self, biased: bool = True) -> float:
        """
        Calculate the skewness of the intensity timeseries.
        The standard deviation is calculated with the biased estimator.

        Parameters
        ----------
        biased_sd : bool, optional
            Whether to use the biased estimator of the skewness, by default True.

        Returns
        -------
        skewness : float
        """
        if self._time.size == 0:
            return np.nan
        return _isnapshot_skew(self._time, self._intensity, biased=biased)

    def pDry(self, threshold: float = 0) -> float:
        """
        Calculate the probability of dryness of the intensity timeseries with a threshold.

        Parameters
        ----------
        threshold : float, optional
            The threshold of dryness, by default 0

        Returns
        -------
        probability of dryness : float
        """
        if self._time.size == 0:
            return np.nan
        return _isnapshot_pDry(self._time, self._intensity, threshold=threshold)

    def rescale(self, scale: Union[float, timedelta], abs_tol: float = 1e-10) -> IndexedSnapshot:
        """
        Rescale the IndexedSnapshot timeseries to a different scale.
        New time index will be divided by the scale. And the intensity will be multiplied by the scale.
        All the time index will be divisible by the scale.
        """
        if isinstance(scale, timedelta):
            scale = scale.total_seconds() / 3600

        if self._time.size == 0:
            return type(self)()
        scale_time, scale_intensity = _isnapshot_rescale(
            self._time, self._intensity, scale, abs_tol
        )
        return type(self)(scale_time, scale_intensity)

    def unpack(self) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        """
        Unpack the IndexedSnapshot timeseries into a normal timeseries.
        """
        if self._time.size == 0:
            return np.array([], dtype=np.float64), np.array([], dtype=np.float64)
        else:
            if not np.all(self._time % 1 == 0):
                print(
                    "Warning: Unpacking IndexedSnapshot with float time index. Resulting time index will be rounded to integer."
                )
            diff_time = np.diff(self._time).astype(np.int64)
            intensity = np.repeat(self._intensity[:-1], diff_time)
            time = np.arange(self._time[0], self._time[-1])
            return time, intensity


def _isnapshot_check(
    time: FloatArray = None,
    intensity: FloatArray = None,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    # Check if both time and intensity are None
    if time is None or intensity is None:
        if time is not None or intensity is not None:
            raise ValueError("time and intensity must both be None or not None")

        _time: npt.NDArray[np.float64] = np.array([], dtype=np.float64)

        _intensity: npt.NDArray[np.float64] = np.array([], dtype=np.float64)

        return _time, _intensity

    time = np.array(time, dtype=np.float64)
    intensity = np.array(intensity, dtype=np.float64)

    # Check if time and intensity have the same length
    if time.size != intensity.size:
        raise ValueError("time and intensity must have the same length")

    # Check if time and intensity are 1D arrays
    if time.ndim != 1 or intensity.ndim != 1:
        raise ValueError("time and intensity must be 1D arrays")

    if time.size < 2:
        raise ValueError(f"time and intensity must have at least 2 elements. Got {time.size}.")

    # Check if the RLE time is strictly increasing.
    if np.any(np.diff(time) <= 0):
        raise ValueError("time must be strictly increasing")
    # Check if there are any nan in the time
    if np.any(np.isnan(time)):
        raise ValueError("time must not contain any nan")
    # Add ending time and intensity if there isn't one

    unique_dt = np.unique(np.diff(time))

    if len(unique_dt) == 0:
        raise ValueError("time must have at least 2 unique values")

    # Calculate count of unique time intervals
    unique_dt_count = np.bincount(np.searchsorted(unique_dt, np.diff(time)))
    # Use the most common time interval as the time interval
    time_interval = unique_dt[np.argmax(unique_dt_count)]

    if not np.isnan(intensity[-1]):
        time = np.append(time, time[-1] + time_interval)
        intensity = np.append(intensity, np.nan)

    # Check if all np.nan are at the end of the intensity timeseries
    # if (num_of_nan := np.sum(np.isnan(intensity))) >= 1:
    #    if not np.all(np.isnan(intensity[-num_of_nan:])):
    #        raise ValueError(
    #            "All np.nan must be at the end of the intensity timeseries"
    #        )

    # if num_of_nan > 1:
    #    time = time[: -num_of_nan + 1]
    #    intensity = intensity[: -num_of_nan + 1]

    # Make sure it's in IndexedSnapshot format
    # intensity_idx = np.diff(intensity, prepend=-np.inf) != 0
    intensity_idx = _index_of_first_appearance_of_consecutive_value(intensity)

    return np.array(time[intensity_idx]), np.array(intensity[intensity_idx])


@nb.njit("i8[:](f8[:])", cache=True)
def _index_of_first_appearance_of_consecutive_value(
    arr: npt.NDArray[np.float64],
) -> npt.NDArray[np.int64]:
    """
    ### This is an internal function. You shouldn't use it directly unless you know what you are doing.

    This function returns the index of the first appearance of each unique value in the array.

    Parameters
    ----------
    arr : npt.NDArray[np.float64]
        1D array of float64 values.

    Returns
    -------
    index : npt.NDArray[np.int64]
        1D array of int64 values.
    """
    is_first_appearance = np.zeros_like(arr, dtype=np.int64)
    is_first_appearance[0] = 1
    last_value = arr[0]
    for i in range(1, len(arr)):
        if np.isnan(last_value) and np.isnan(arr[i]):
            continue

        if np.isnan(last_value) and not np.isnan(arr[i]):
            is_first_appearance[i] = 1
            last_value = arr[i]
            continue

        if arr[i] != last_value:
            is_first_appearance[i] = 1
            last_value = arr[i]

    first_appearance_idx = np.where(is_first_appearance == 1)[0]

    return first_appearance_idx


def _intensity_delta_check(
    time: npt.NDArray[np.float64],
    intensity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    if len(time) != 0:
        return np.column_stack((time, np.diff(intensity[:-1], prepend=0, append=0)))
    else:
        return np.array([], dtype=np.float64)


def _delta_to_isnapshot_archived(
    time: npt.NDArray[np.float64],
    intensity_delta: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    ### This is an internal function. You shouldn't use it directly unless you know what you are doing.

    This function convert the delta encoding of the intensity timeseries into **ALMOST** IndexedSnapshot format.
    The np.nan at the end of the intensity timeseries is not included in the IndexedSnapshot format.

    Parameters
    ----------
    time : npt.NDArray[np.float64]
        1D array of time index that follows delta encoding format.
    intensity_delta : npt.NDArray[np.float64]
        1D array of intensity that follows delta encoding format.

    Returns
    -------
    time : npt.NDArray[np.float64]
        1D array of time index that follows IndexedSnapshot format (Without the np.nan at the end).
    intensity : npt.NDArray[np.float64]
        1D array of intensity that follows IndexedSnapshot format (Without the np.nan at the end).
    """

    # Zip time and intensity_delta into a 2D array
    delta_encoding = np.column_stack((time, intensity_delta))

    # Sort the change_time_idx by time
    delta_encoding = delta_encoding[np.argsort(delta_encoding[:, 0])]

    # Calculate the cumulative sum of the intensity changes inplace.
    np.cumsum(delta_encoding[:, 1], out=delta_encoding[:, 1])

    ## TODO: Round to 10 decimal places to avoid floating point errors. And it's inplace.
    delta_encoding.round(10, out=delta_encoding)

    # Remove duplicate times by keeping the last occurence
    # Since the np.unique function only keeps the first occurence,
    # We reverse the array, keep the first occurence, then reverse it back.
    _, unique_idx = np.unique(delta_encoding[:, 0][::-1], return_index=True)
    # Adjust indices to account for reversed order
    unique_indices = len(delta_encoding) - 1 - unique_idx

    return (
        delta_encoding[:, 0][unique_indices],
        delta_encoding[:, 1][unique_indices],
    )


# Signature of two float64 arrays input and two float64 arrays output
@nb.njit("UniTuple(f8[:], 2)(f8[:], f8[:])", cache=True)
def _delta_to_isnapshot(
    time: npt.NDArray[np.float64],
    intensity_delta: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    ### This is an internal function. You shouldn't use it directly unless you know what you are doing.

    This function convert the delta encoding of the intensity timeseries into **ALMOST** IndexedSnapshot format.
    The np.nan at the end of the intensity timeseries is not included in the IndexedSnapshot format.

    Parameters
    ----------
    time : npt.NDArray[np.float64]
        1D array of time index that follows delta encoding format.
    intensity_delta : npt.NDArray[np.float64]
        1D array of intensity that follows delta encoding format.

    Returns
    -------
    time : npt.NDArray[np.float64]
        1D array of time index that follows IndexedSnapshot format (Without the np.nan at the end).
    intensity : npt.NDArray[np.float64]
        1D array of intensity that follows IndexedSnapshot format (Without the np.nan at the end).
    """

    # Sort the change_time_idx by time
    sorted_idx = np.argsort(time)
    time = time[sorted_idx]
    intensity_delta = intensity_delta[sorted_idx]

    # Calculate the cumulative sum of the intensity changes inplace.
    for i in range(1, len(intensity_delta)):
        intensity_delta[i] += intensity_delta[i - 1]

    intensity_delta[np.where(np.abs(intensity_delta) < 1e-10)] = 0

    # Remove duplicate times by keeping the last occurence
    diff_idx = np.empty(len(time), dtype=np.bool_)
    for i in range(len(time) - 1, 0, -1):
        diff_idx[i - 1] = time[i] != time[i - 1]
    diff_idx[-1] = True

    return (
        time[diff_idx],
        intensity_delta[diff_idx],
    )


def _merge_isnapshot(a: IndexedSnapshot, b: IndexedSnapshot) -> IndexedSnapshot:
    """
    Merge two IndexedSnapshot timeseries together.

    Parameters
    ----------
    a : IndexedSnapshot
        The first IndexedSnapshot timeseries.
    b : IndexedSnapshot
        The second IndexedSnapshot timeseries.

    Returns
    -------
    IndexedSnapshot
        The merged IndexedSnapshot timeseries.
    """
    if len(a.intensity_delta) == 0 and len(b.intensity_delta) == 0:
        return IndexedSnapshot()
    if len(a.intensity_delta) == 0:
        intensity_delta = b.intensity_delta
    elif len(b.intensity_delta) == 0:
        intensity_delta = a.intensity_delta
    else:
        intensity_delta = np.concatenate((a.intensity_delta, b.intensity_delta))

    time, intensity = _delta_to_isnapshot(intensity_delta[:, 0], intensity_delta[:, 1])
    # The intensity at the end of the timeseries should be np.nan
    # But when we extract the intensity_delta, we replace it with 0 to make the calculation easier.
    # Other 0 value is included in the timeseries.
    # But the last 0 value is not included in the timeseries. So we need to add it back by replacing the last value with np.nan
    intensity[-1] = np.nan

    if len(time) == 0:
        return IndexedSnapshot()

    return IndexedSnapshot(time, intensity)


@nb.njit("UniTuple(f8, 2)(f8[:], f8[:])", cache=True)
def _isnapshot_sum_and_duration(
    time: npt.NDArray[np.float64], intensity: npt.NDArray[np.float64]
) -> tuple[float, float]:
    """
    ### This is an internal function. Use `IndexedSnapshot.mean()` instead.

    Calculate the total depth of the intensity timeseries.

    Parameters
    ----------
    time : npt.NDArray[np.float64]
        1D array of time index that follows IndexedSnapshot format.
    intensity : npt.NDArray[np.float64]
        1D array of intensity that follows IndexedSnapshot format.

    Returns
    -------
    mean: float
        The total depth of the intensity timeseries.

    """
    # each_intensity_duration = np.diff(time)
    depth: float = 0.0
    duration: float = 0.0
    for i in range(len(time) - 1):
        if np.isnan(intensity[i]):
            continue
        depth += intensity[i] * (time[i + 1] - time[i])
        duration += time[i + 1] - time[i]

    return depth, duration


@nb.njit("f8(f8[:], f8[:])", cache=True)
def _isnapshot_mean(
    time: npt.NDArray[np.float64], intensity: npt.NDArray[np.float64]
) -> float:
    """
    ### This is an internal function. Use `IndexedSnapshot.mean()` instead.

    Calculate the mean of the intensity timeseries.

    Parameters
    ----------
    time : npt.NDArray[np.float64]
        1D array of time index that follows IndexedSnapshot format.
    intensity : npt.NDArray[np.float64]
        1D array of intensity that follows IndexedSnapshot format.

    Returns
    -------
    mean: float
        The mean of the intensity timeseries.

    """
    depth, duration = _isnapshot_sum_and_duration(time, intensity)

    if duration == 0:
        return np.nan

    return depth / duration


@nb.njit("f8(f8[:], f8[:])", cache=True)
def _isnapshot_sum_squared_error(
    time: npt.NDArray[np.float64],
    intensity: npt.NDArray[np.float64],
) -> float:
    """
    ### This is an internal function. Use `IndexedSnapshot.standard_deviation()` instead.

    Calculate the sum of squared error of the intensity timeseries.

    Parameters
    ----------
    time : npt.NDArray[np.float64]
        1D array of time index that follows IndexedSnapshot format.
    intensity : npt.NDArray[np.float64]
        1D array of intensity that follows IndexedSnapshot format.

    Returns
    -------
    sum of squared error: float
        The sum of squared error of the intensity timeseries.
    """
    mean = _isnapshot_mean(time, intensity)
    sse = 0
    for i in range(len(time) - 1):
        if np.isnan(intensity[i]):
            continue
        sse += (intensity[i] - mean) ** 2 * (time[i + 1] - time[i])
    return sse


@nb.njit("f8(f8[:], f8[:], b1)", cache=True)
def _isnapshot_variance(
    time: npt.NDArray[np.float64],
    intensity: npt.NDArray[np.float64],
    biased: bool = False,
) -> float:
    """
    ### This is an internal function. Use `IndexedSnapshot.standard_deviation()` instead.

    Calculate the variance of the intensity timeseries.

    Parameters
    ----------
    time : npt.NDArray[np.float64]
        1D array of time index that follows IndexedSnapshot format.
    intensity : npt.NDArray[np.float64]
        1D array of intensity that follows IndexedSnapshot format.

    Returns
    -------
    variance: float
        The variance of the intensity timeseries.
    """
    depth, duration = _isnapshot_sum_and_duration(time, intensity)
    sse = _isnapshot_sum_squared_error(time, intensity)

    if duration - (biased is False) == 0:
        return np.nan

    return sse / (duration - (biased is False))


@nb.njit("f8(f8[:], f8[:], b1)", cache=True)
def _isnapshot_standard_deviation(
    time: npt.NDArray[np.float64],
    intensity: npt.NDArray[np.float64],
    biased: bool = False,
) -> float:
    """
    ### This is an internal function. Use `IndexedSnapshot.standard_deviation()` instead.

    Calculate the standard deviation of the intensity timeseries.

    Parameters
    ----------
    time : npt.NDArray[np.float64]
        1D array of time index that follows IndexedSnapshot format.
    intensity : npt.NDArray[np.float64]
        1D array of intensity that follows IndexedSnapshot format.

    Returns
    -------
    standard deviation: float
        The standard deviation of the intensity timeseries.
    """
    return _isnapshot_variance(time, intensity, biased) ** 0.5


@nb.njit("f8(f8[:], f8[:], b1)", cache=True)
def _isnapshot_coef_var(
    time: npt.NDArray[np.float64],
    intensity: npt.NDArray[np.float64],
    biased: bool = False,
) -> float:
    """
    ### This is an internal function. Use `IndexedSnapshot.standard_deviation()` instead.

    Calculate the coefficient of variation of the intensity timeseries.

    Parameters
    ----------
    time : npt.NDArray[np.float64]
        1D array of time index that follows IndexedSnapshot format.
    intensity : npt.NDArray[np.float64]
        1D array of intensity that follows IndexedSnapshot format.
    mean : float, optional
        The mean of the intensity timeseries, If None, it will be calculated, by default None
    sse : float, optional
        The sum of squared error of the intensity timeseries, If None, it will be calculated, by default None
    biased : bool, optional
        Whether to use the biased estimator of the standard deviation, by default False.

    Returns
    -------
    coefficient of variation: float
        The coefficient of variation of the intensity timeseries.

    """
    # Coefficient of variation
    mean = _isnapshot_mean(time, intensity)

    if np.isclose(mean, 0, atol=1e-15):
        return np.nan

    return _isnapshot_standard_deviation(time, intensity, biased) / mean


@nb.njit("f8(f8[:], f8[:], b1)", cache=True)
def _isnapshot_skew(
    time: npt.NDArray[np.float64],
    intensity: npt.NDArray[np.float64],
    biased: bool = False,
) -> float:
    """
    ### This is an internal function. Use `IndexedSnapshot.standard_deviation()` instead.

    Calculate the skewness of the intensity timeseries.
    The standard deviation is calculated with the unbiased estimator.

    Parameters
    ----------
    time : npt.NDArray[np.float64]
        1D array of time index that follows IndexedSnapshot format.
    intensity : npt.NDArray[np.float64]
        1D array of intensity that follows IndexedSnapshot format.

    Returns
    -------
    skewness : float
        The skewness of the intensity timeseries.
    """
    # Skewness
    depth, duration = _isnapshot_sum_and_duration(time, intensity)

    if duration == 0:
        return np.nan

    mean = depth / duration

    stddev = _isnapshot_standard_deviation(time, intensity, biased=True)
    if np.isclose(stddev, 0, atol=1e-15):
        return np.nan

    skewness = 0
    for i in range(len(time) - 1):
        if np.isnan(intensity[i]):
            continue
        skewness += (intensity[i] - mean) ** 3 * (time[i + 1] - time[i])
    skewness = skewness / (duration * stddev**3)

    if biased is False:
        skewness *= (duration * (duration - 1)) ** 0.5 / (duration - 2)
    return skewness


def _isnapshot_pDry(
    time: npt.NDArray[np.float64],
    intensity: npt.NDArray[np.float64],
    threshold: float = 0,
) -> float:
    """
    ### This is an internal function. Use `IndexedSnapshot.pDry()` instead.

    Calculate the probability of dryness of the intensity timeseries with a threshold.

    Parameters
    ----------
    time : npt.NDArray[np.float64]
        1D array of time index that follows IndexedSnapshot format.
    intensity : npt.NDArray[np.float64]
        1D array of intensity that follows IndexedSnapshot format.
    threshold : float, optional
        The threshold of dryness, by default 0

    Returns
    -------
    probability of dryness : float
    """
    duration = 0
    wet_duration = 0
    for i in range(len(time) - 1):
        if np.isnan(intensity[i]):
            continue
        duration += time[i + 1] - time[i]
        if intensity[i] > threshold:
            wet_duration += time[i + 1] - time[i]

    if duration == 0:
        return np.nan

    return 1 - (wet_duration / duration)


@nb.njit("f8(f8[:], f8[:], f8)", cache=True)
def _isnapshot_acf(
    time: npt.NDArray[np.float64],
    intensity: npt.NDArray[np.float64],
    lag: Optional[float] = 1,
):
    """
    ### This is an internal function. Use `IndexedSnapshot.acf()` instead.

    Calculate the n-lag autocorrelation coefficient of the intensity timeseries.
    It's basically Cov(X_t, X_{t+lag}) / Var(X_t)

    Parameters
    ----------
    time : npt.NDArray[np.float64]
        1D array of time index that follows IndexedSnapshot format.
    intensity : npt.NDArray[np.float64]
        1D array of intensity that follows IndexedSnapshot format.
    lag : int, optional
        The lag of the autocorrelation coefficient, by default 1

    Returns
    -------
    n-lag Autocorrelation coefficient: float
        The n-lag autocorrelation coefficient of the intensity timeseries.
    """
    if lag == 0:
        return 1
    n = len(time)

    mean = _isnapshot_mean(time, intensity)

    sse = _isnapshot_sum_squared_error(time, intensity)

    if np.isclose(sse, 0, atol=1e-15):
        return np.nan

    shift = 0
    for time_idx in time:
        if time_idx < lag + time[0]:
            shift += 1
        else:
            break

    x_idx = shift
    y_idx = 1
    result = 0
    while x_idx < n:
        if time[y_idx] >= time[x_idx] - lag:
            if np.isnan(intensity[x_idx - 1]) or np.isnan(intensity[y_idx - 1]):
                x_idx += 1
                continue
            size = (time[x_idx] - lag) - max(time[y_idx - 1], (time[x_idx - 1] - lag))
            result += (
                (intensity[y_idx - 1] - mean) * (intensity[x_idx - 1] - mean)
            ) * size
            x_idx += 1
        else:
            if np.isnan(intensity[x_idx - 1]) or np.isnan(intensity[y_idx - 1]):
                y_idx += 1
                continue
            size = time[y_idx] - max(time[y_idx - 1], (time[x_idx - 1] - lag))
            result += (
                (intensity[y_idx - 1] - mean) * (intensity[x_idx - 1] - mean)
            ) * size
            y_idx += 1

    return result / sse

@nb.njit("f8(f8, f8)", cache=True)
def mod(a: float, b: float) -> float:
    return a - b * np.floor(a / b)

@nb.njit("UniTuple(f8[:], 2)(f8[:], f8[:], f8, f8)", cache=True)
def _isnapshot_rescale(
    time: npt.NDArray[np.float64], intensity: npt.NDArray[np.float64], scale: float, abs_tol: float = 1e-10
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    ### This is an internal function. Use `IndexedSnapshot.standard_deviation()` instead.

    Rescale the IndexedSnapshot timeseries to a different scale.
    New time index will be divided by the scale. And the intensity will be multiplied by the scale.
    All the time index will be divisible by the scale.

    Parameters
    ----------
    time : npt.NDArray[np.float64]
        1D array of time index that follows IndexedSnapshot format.
    intensity : npt.NDArray[np.float64]
        1D array of intensity that follows IndexedSnapshot format.
    scale : float
        The scale to be rescaled to.
    abs_tol : float, optional
        The absolute tolerance for floating point error, by default 1e-10

    Returns
    -------
    scaled time : npt.NDArray[np.float64]
        1D array of time index that follows IndexedSnapshot format.
    scaled intensity : npt.NDArray[np.float64]
        1D array of intensity that follows IndexedSnapshot format.

    Examples
    --------
    >>> time = np.array([0, 10, 13, 18, 33])
    >>> intensity = np.array([0, 3, 0, 5, np.nan])
    >>> scale = 5
    >>> _isnapshot_rescale(time, intensity, scale)
    (array([0., 2., 3., 4., 6., 7.]), array([ 0.,  9., 10., 25., 15., nan]))
    """
    time_index = len(time) - 1
    scale_time = np.full(time_index * 3 + 2, np.nan)
    scale_intensity = np.full(time_index * 3 + 2, np.nan)
    #scale_time = np.zeros(time_index * 3 + 2)
    #scale_intensity = np.zeros(time_index * 3 + 2)

    scale_time[0] = np.nan
    rescale_idx = 1
    for i in range(time_index):
        srt, end = time[i], time[i + 1]
        r_srt, r_end = np.floor(srt / scale), np.floor(end / scale)
        intensity_i = intensity[i]

        # Case 1 might overshot. Case 3 and Case 2 almost always overshot. As you can see in their blow diagram.
        # But we still move to the next rescale time index. So we need to drop back to the previous rescale time index.
        if scale_time[rescale_idx - 1] == r_srt:
            rescale_idx -= 1

        # Case 1: When two time index are in the same rescale range.
        # rescale:      |-------------------------------------|
        #                   |  A  |  X  |  X  |  ...
        # original:         |-----|-----|-----|  ...
        if r_srt == r_end:
            scale_time[rescale_idx] = r_srt
            if np.isnan(intensity_i):
                pass
            elif np.isnan(scale_intensity[rescale_idx]):
                scale_intensity[rescale_idx] = intensity_i * (end - srt)
            else:
                scale_intensity[rescale_idx] += intensity_i * (end - srt)
            rescale_idx += 1
            continue

        # Case 2: When two time index cross "1" rescale range.
        # rescale:    |-----------|-----------|
        #                 |   A   |  B  |
        # original:       |-------------|
        if r_srt + 1 == r_end:
            scale_time[rescale_idx] = r_srt
            if np.isnan(intensity_i):
                pass
            elif np.isnan(scale_intensity[rescale_idx]):
                scale_intensity[rescale_idx] = intensity_i * (scale - mod(srt, scale))
            else:
                scale_intensity[rescale_idx] += intensity_i * (scale - mod(srt, scale))
            rescale_idx += 1

            if np.isclose(mod(end, scale), 0, atol=1e-10):
                # It means that the end time is the same as the next rescale time.
                # We shouldn't add the intensity 0 to the next rescale time.
                # Because it might make nan intensity to be 0. Which is wrong.
                continue

            scale_time[rescale_idx] = r_end
            if np.isnan(intensity_i):
                pass
            elif np.isnan(scale_intensity[rescale_idx]):
                scale_intensity[rescale_idx] = intensity_i * mod(end, scale)
            else:
                scale_intensity[rescale_idx] += intensity_i * mod(end, scale)
            rescale_idx += 1
            continue

        # Case 3: When two time index cross multiple rescale range. We can finish all the middle rescale range (B).
        # rescale:    |-----------|------------|-----------|
        #                     | A |     B      | C |
        # original:           |--------------------|
        if r_srt + 1 < r_end:
            scale_time[rescale_idx] = r_srt
            if np.isnan(intensity_i):
                pass
            elif np.isnan(scale_intensity[rescale_idx]):
                scale_intensity[rescale_idx] = intensity_i * (scale - mod(srt, scale))
            else:
                scale_intensity[rescale_idx] += intensity_i * (scale - mod(srt, scale))
            rescale_idx += 1

            scale_time[rescale_idx] = r_srt + 1
            if np.isnan(intensity_i):
                pass
            elif np.isnan(scale_intensity[rescale_idx]):
                scale_intensity[rescale_idx] = intensity_i * scale
            else:
                scale_intensity[rescale_idx] += intensity_i * scale
            rescale_idx += 1

            if np.isclose(mod(end, scale), 0, atol=1e-10):
                # It means that the end time is the same as the next rescale time.
                # We shouldn't add the intensity 0 to the next rescale time.
                # Because it might make nan intensity to be 0. Which is wrong.
                continue

            scale_time[rescale_idx] = r_end
            if np.isnan(intensity_i):
                pass
            elif np.isnan(scale_intensity[rescale_idx]):
                scale_intensity[rescale_idx] = intensity_i * mod(end, scale)
            else:
                scale_intensity[rescale_idx] += intensity_i * mod(end, scale)
            rescale_idx += 1
            continue

    # Append the ending time with nan intensity
    # If the last rescaled time is the same as the last original time, change it to nan.

    scale_time = scale_time[1 : rescale_idx + 1]
    scale_time[-1] = np.ceil(time[-1] / scale)
    scale_intensity = scale_intensity[1 : rescale_idx + 1]

    # Preallocate boolean array
    diff_time = np.ones(len(scale_time), dtype=np.bool_)
    diff_time[0] = True
    #round_digit = int(-np.log10(abs_tol))
    for i in range(1, len(scale_time)):
        # Skip current time index if it's the same as the previous one.
        ## Current intensity is the same as the previous one.
        if scale_intensity[i] == scale_intensity[i - 1]:
            diff_time[i] = False
            continue
        ## Current intensity is close to the previous one. (Floating point error)
        if np.isclose(scale_intensity[i], scale_intensity[i - 1], rtol=abs_tol):
            diff_time[i] = False
            continue
        ## Current intensity is the same as the previous one. (Both are np.nan)
        if np.isnan(scale_intensity[i]) and np.isnan(scale_intensity[i - 1]):
            diff_time[i] = False
            continue

        # Keep the current time index if it's different from the previous one.
        ## Current intensity is np.nan.
        if np.isnan(scale_intensity[i]):
            continue

        ## Current intensity is float. Round it to atol.
        #scale_intensity[i] = np.round(scale_intensity[i], round_digit)

    # Replace all np.nan in scale_intensity to 0
    #scale_intensity[np.isnan(scale_intensity)] = 0

    return scale_time[diff_time], scale_intensity[diff_time]



from __future__ import annotations

import numpy as np
import numpy.typing as npt


class Storm:
    def __init__(self, start: float, duration: float, cells: npt.NDArray[np.float64], eta: float, mux: float, gamma: float, beta: float):
        self.start = start
        self.duration = duration
        self.cells = cells
        self.eta = eta
        self.mux = mux
        self.gamma = gamma
        self.beta = beta




from __future__ import annotations

import math
import warnings
from dataclasses import dataclass
from datetime import timedelta
from typing import Any, Callable, List, Optional, Tuple, Union

import numba as nb  # type: ignore
import numpy as np
import numpy.typing as npt
import pandas as pd  # type: ignore
import scipy as sp  # type: ignore

from pybl.models import BaseBLRP, BaseBLRP_RCIModel, StatMetrics
from pybl.raincell import ExponentialRCIModel, IConstantRCI, Storm
from pybl.timeseries import IndexedSnapshot


@dataclass
class BLRPRx_params:
    # Kaczmarska et al. (2014)
    lambda_: float  # Storm's arrival rate (per hour) parameter of Poisson distribution
    phi: float  # ratio of the storm (cell process) termination rate to η (i.e. γ/η)j
    kappa: float  # ratio of the cell arrival rate to η (i.e. β/η)
    alpha: float  # shape parameter for the Gamma distribution of the cell duration parameter, η
    nu: float  # scale parameter for the Gamma distribution of the cell duration parameter, η
    sigmax_mux: float  # ratio of the standard deviation of cell intensity to the mean cell intensity
    iota: float  # ratio of mean cell intensity to η (i.e. µ_x /η)

    def size(self):
        return 7

    def unpack(
        self,
    ) -> Tuple[
        float,
        float,
        float,
        float,
        float,
        float,
        float,
    ]:
        return (
            self.lambda_,
            self.phi,
            self.kappa,
            self.alpha,
            self.nu,
            self.sigmax_mux,
            self.iota,
        )

    def copy(self) -> BLRPRx_params:
        return BLRPRx_params(*self.unpack())


class BLRPRx(BaseBLRP):
    ## This BLRPRx is an implementation of Kaczmarska et al. (2014)
    __slots__ = (
        "rci_model",
        "rng",
        "params",
    )

    def __init__(
        self,
        params: Optional[BLRPRx_params] = None,
        sampling_rng: Optional[np.random.Generator] = None,
        rci_model: BaseBLRP_RCIModel = None,
    ) -> None:
        super().__init__(sampling_rng, rci_model)
        # If user does not provide params, give the default values.
        if params is None:
            self.params = BLRPRx_params(
                lambda_=5.0,
                phi=5.0,
                kappa=5.0,
                alpha=5.0,
                nu=5.0,
                sigmax_mux=5.0,
                iota=5.0,
            )
        else:
            self.params = params.copy()


    def copy(self, rng: Optional[np.random.Generator] = None) -> BLRPRx:
        if rng is None:
            rng = np.random.default_rng(self.rng)

        return BLRPRx(
            params=self.get_params(),
            sampling_rng=rng,
            rci_model=type(self.rci_model)(),
        )

    def get_params(self) -> BLRPRx_params:
        return self.params.copy()

    def update_params(self, params: BLRPRx_params) -> None:
        self.params = params.copy()

    def _kernel(self, k: float, u: float, nu: float, alpha: float) -> float:
        return _blrprx_kernel(k, u, nu, alpha)

    def mean(self, timescale: float = 1.0) -> float:
        return _blrprx_mean(timescale, *self.get_params().unpack())

    def variance(self, timescale: float = 1.0) -> float:
        f1 = self.rci_model.get_f1(sigmax_mux=self.params.sigmax_mux)
        return _blrprx_variance(timescale, f1, *self.get_params().unpack())

    def covariance(self, timescale: float = 1.0, lag: float = 1.0) -> float:
        f1 = self.rci_model.get_f1(sigmax_mux=self.params.sigmax_mux)
        return _blrprx_covariance(timescale, f1, lag, *self.get_params().unpack())

    def moment_3rd(self, timescale: float = 1.0) -> float:
        f1 = self.rci_model.get_f1(sigmax_mux=self.params.sigmax_mux)
        f2 = self.rci_model.get_f2(sigmax_mux=self.params.sigmax_mux)
        return _blrprx_moment_3rd(timescale, f1, f2, *self.get_params().unpack())

    def get_stats(
        self,
        stat_metric: StatMetrics,
        timescale: Union[float, timedelta] = 1.0,
    ) -> float:
        if isinstance(timescale, timedelta):
            timescale = timescale.total_seconds() / 3600.0

        if stat_metric == StatMetrics.MEAN:
            return self.mean(timescale)
        elif stat_metric == StatMetrics.VAR:
            return self.variance(timescale)
        elif stat_metric == StatMetrics.CVAR:
            return np.sqrt(self.variance(timescale)) / self.mean(timescale)
        elif stat_metric == StatMetrics.AR1:
            return self.covariance(timescale, 1.0) / self.variance(timescale)
        elif stat_metric == StatMetrics.AR2:
            return self.covariance(timescale, 2.0) / self.variance(timescale)
        elif stat_metric == StatMetrics.AR3:
            return self.covariance(timescale, 3.0) / self.variance(timescale)
        elif stat_metric == StatMetrics.AC1:
            return self.covariance(timescale, 1.0)
        elif stat_metric == StatMetrics.AC2:
            return self.covariance(timescale, 2.0)
        elif stat_metric == StatMetrics.AC3:
            return self.covariance(timescale, 3.0)
        elif stat_metric == StatMetrics.SKEWNESS:
            return self.moment_3rd(timescale) / np.power(self.variance(timescale), 1.5)
        else:
            # Not implemented properties
            return 0.0
    def get_stats_dataframe(
        self,
        stat_metrics: list[StatMetrics],
        timescales: Union[list[float], list[timedelta]] = [1.0],
    ) -> pd.DataFrame:
        for scale_idx, scale in enumerate(timescales):
            if not (isinstance(scale, float) or isinstance(scale, timedelta)):
                raise ValueError("timescales must be a list of float or timedelta")

        stats_arr = np.empty((len(timescales), len(stat_metrics)), dtype=np.float64)
        for scale_idx, scale in enumerate(timescales):

            if isinstance(scale, timedelta):
                float_scale = scale.total_seconds() / 3600.0
            elif isinstance(scale, float):
                float_scale = scale

            for prop_idx, prop in enumerate(stat_metrics):
                stats_arr[scale_idx, prop_idx] = self.get_stats(prop, float_scale)

        stats_df = pd.DataFrame(
            stats_arr, columns=stat_metrics, index=timescales
        )
        stats_df.index.name = "timescale_hr"

        return stats_df

    def sample_raw(self, duration_hr: float) -> npt.NDArray[np.float64]:
        lambda_, phi, kappa, alpha, nu, sigmax_mux, iota = self.get_params().unpack()
        rng = self.rng

        # Storm number sampling
        n_storm = rng.poisson(lambda_ * duration_hr)

        # Calculate the original parameters
        eta = rng.gamma(alpha, 1 / nu, size=n_storm)  # (n_storm, )
        gamma = phi * eta  # (n_storm, )
        beta = kappa * eta  # (n_storm, )
        mux = iota * eta  # (n_storm, )

        # Storm parameters sampling
        storm_starts = rng.uniform(0, duration_hr, n_storm)  # (n_storm, )
        storm_durations = rng.exponential(1 / gamma, n_storm)  # (n_storm, )

        # Cell sampling
        n_cells_per_storm = 1 + rng.poisson(
            beta * storm_durations, size=n_storm
        )  # (n_storm, )
        total_cells: int = n_cells_per_storm.sum()

        # Pre-allocate arrays
        cell_starts = np.zeros(total_cells)  # (total_cells, )
        cell_durations = np.zeros(total_cells)  # (total_cells, )
        cell_intensities = np.zeros(total_cells)  # (total_cells, )

        cells_start_idx = 0
        for i, (s, d) in enumerate(zip(storm_starts, storm_durations)):
            cell_starts[cells_start_idx] = (
                s  # First cell starts at the same time as the storm
            )
            cell_starts[
                cells_start_idx + 1 : cells_start_idx + n_cells_per_storm[i]
            ] = rng.uniform(s, s + d, n_cells_per_storm[i] - 1)

            cell_durations[cells_start_idx : cells_start_idx + n_cells_per_storm[i]] = (
                rng.exponential(scale=1 / eta[i], size=n_cells_per_storm[i])
            )

            cell_intensities[
                cells_start_idx : cells_start_idx + n_cells_per_storm[i]
            ] = self.rci_model.sample_intensity(
                rng=rng, mux=mux[i], sigmax_mux=sigmax_mux, n_cells=n_cells_per_storm[i]
            )
            cells_start_idx += n_cells_per_storm[i]

        cell_ends = cell_starts + cell_durations  # (total_cells, )

        # Flatten cell_starts, cell_ends, cell_intensities and stack them together
        cell_arr = np.stack(
            (cell_starts, cell_ends, cell_intensities), axis=-1
        )  # (total_cells, 3)

        return cell_arr

    def sample(self, duration_hr: float) -> IndexedSnapshot:
        # cell_arr = self.sample_raw(duration_hr)
        cell_arr, n_cells_per_storm, storms_info = _blrprx_sample(
            duration_hr,
            self.rci_model.sample_intensity,
            self.rng,
            *self.get_params().unpack(),
        )

        delta = np.concatenate(
            [
                np.stack([cell_arr[:, 0], cell_arr[:, 2]]).T,
                np.stack([cell_arr[:, 1], -cell_arr[:, 2]]).T,
            ],
            axis=0,
        )
        ts = IndexedSnapshot.fromDelta(time=delta[:, 0], intensity_delta=delta[:, 1])

        return ts[0:duration_hr]  # type: ignore

    def sample_storms(self, duration_hr: float) -> Tuple[IndexedSnapshot, List[Storm]]:
        cell_arr, n_cells_per_storm, storms_info = _blrprx_sample(
            duration_hr,
            self.rci_model.sample_intensity,
            self.rng,
            *self.get_params().unpack(),
        )

        delta = np.concatenate(
            [
                np.stack([cell_arr[:, 0], cell_arr[:, 2]]).T,
                np.stack([cell_arr[:, 1], -cell_arr[:, 2]]).T,
            ],
            axis=0,
        )
        ts = IndexedSnapshot.fromDelta(time=delta[:, 0], intensity_delta=delta[:, 1])

        storms = []

        cum_cells = 0
        for i, storm_info in enumerate(storms_info):
            storms.append(
                Storm(
                    cells=cell_arr[cum_cells : cum_cells + n_cells_per_storm[i]],
                    start=storm_info[0],
                    duration=storm_info[1],
                    eta=storm_info[2],
                    mux=storm_info[3],
                    gamma=storm_info[4],
                    beta=storm_info[5],
                )
            )
            cum_cells += n_cells_per_storm[i]

        return ts[0:duration_hr], storms  # type: ignore

    def fit(
        self,
        stats: pd.DataFrame,
        weight: pd.DataFrame,
        rng: Optional[np.random.Generator] = None,
        tol: float = 0.5,
    ) -> dict[str, Any]:

        warnings.filterwarnings("ignore", category=sp.optimize.OptimizeWarning)


        if rng is None:
            rng = np.random.default_rng()
        bound = self.bound_estimation(stats, weight)

        fitter = BLRPRxConfig(stats, weight, self.rci_model)
        obj = fitter.get_evaluation_func()

        for i in range(20):
            if obj(np.array(self.params.unpack())) < tol:
                break

            # Clip the parameters to the bound

            guess = np.clip(np.array(self.params.unpack()), [b[0] for b in bound], [b[1] for b in bound])

            result_bh = sp.optimize.basinhopping(
                obj,
                x0=guess,
                T=10,
                stepsize=1,
                niter=20,
                minimizer_kwargs={"method": "Nelder-Mead", "bounds": bound},
                seed=rng,
            )

            if obj(np.array(self.params.unpack())) > result_bh.fun:
                self.update_params(BLRPRx_params(*result_bh.x))

        if obj(np.array(self.params.unpack())) < tol:
            status = "Success"
        else:
            status = "Maximum iteration reached"

        stats_metrics: list[StatMetrics] = stats.columns.to_list() # type: ignore
        timescales = stats.index.to_list()

        report = {
            "obj": obj,
            "fun": obj(np.array(self.params.unpack())),
            "x": self.params.unpack(),
            "status": status,
            "rci_model": self.rci_model.__class__.__name__,
            "theo_stats": self.get_stats_dataframe(stat_metrics=stats_metrics, timescales=timescales),
        }

        return report
        # if obj(np.array(self.params.unpack())) < result_bh.fun:
        #    self.update_params(BLRPRx_params(*result_bh.x))

    def bound_estimation(
        self, target: pd.DataFrame, weight: pd.DataFrame
    ) -> List[Tuple[float, float]]:
        bound = [
            (0.0001, 10.0),
            (0.0001, 10.0),
            (0.0001, 10.0),
            (0.0001, 10.0),
            (0.0001, 10.0),
            (0.999, 1.0),
            (0.0001, 10.0),
        ]
        return bound


class BLRPRxConfig:
    __slots__ = ("_weight", "_stats", "_rci_model")

    _weight: pd.DataFrame
    _stats: pd.DataFrame
    _rci_model: IConstantRCI

    def __init__(
        self,
        stats: pd.DataFrame,
        weight: pd.DataFrame,
        rci_model: Optional[IConstantRCI] = None,
    ):
        if not (stats.shape == weight.shape):
            raise ValueError("stats and weight must have the same shape")

        # Check if they have the same set of column regardless of order
        if not (set(stats.columns) == set(weight.columns)):
            raise ValueError("stats and weight must have the same set of columns")

        # Check if they have the same set of index regardless of order
        if not (set(stats.index) == set(weight.index)):
            raise ValueError("stats and weight must have the same set of index")

        # Make sure there isn't duplicate columns
        if len(stats.columns) != len(set(stats.columns)) or len(weight.columns) != len(
            set(weight.columns)
        ):
            raise ValueError("stats and weight must not have duplicate columns")

        # Make sure there isn't duplicate index
        if len(stats.index) != len(set(stats.index)) or len(weight.index) != len(
            set(weight.index)
        ):
            raise ValueError("stats and weight must not have duplicate index")

        # Check if rci_model is None or IConstantRCI
        if rci_model is None:
            self._rci_model = ExponentialRCIModel()
        elif isinstance(rci_model, IConstantRCI):
            self._rci_model = rci_model
        else:
            raise ValueError("rci_model must follow IConstantRCI protocol")

        self._stats = stats.copy(deep=True)
        self._weight = weight.copy(deep=True)

    def get_evaluation_func(self) -> Callable[[npt.NDArray[np.float64]], np.float64]:
        """
        Return a objective function based on your configuration of stats, weight and mask.
        It is a function that takes in a numpy array of 7 parameters and return a score.

        Note that the first time this function is called, it will take a while to compile.

        Returns:
        --------
        evaluation_func: Callable[[npt.NDArray[np.float64]], np.float64]
            A function that takes in a numpy array of 7 parameters and return a error score.
        """
        scales = self._stats.index.to_numpy()
        stats_types_enum = self._stats.columns.values
        for i, stat in enumerate(stats_types_enum):
            if stat not in StatMetrics:
                raise ValueError(f"Invalid StatMetrics: {stat}")
        stats_types = np.array([stat.value for stat in stats_types_enum])

        stats_np = self._stats.to_numpy()
        weight_np = self._weight.to_numpy()

        f1_func = self._rci_model.get_f1
        f2_func = self._rci_model.get_f2

        @nb.njit
        def evaluation_func(x: npt.NDArray[np.float64]) -> np.float64:
            # fmt: off
            (_l ,_p ,_k ,_a ,_n ,_s ,_i) = x  # noqa: E741.   lambda, phi, kappa, alpha, nu, sigma, iota

            # If phi is 1. Many of the formula will have division by 0.
            if _p == 1.0:
                return np.float64(np.nan)

            # fmt: on
            f1 = f1_func(_s)
            f2 = f2_func(_s)
            score = np.float64(0)
            for t, scale in enumerate(scales):
                mean = _blrprx_mean(scale, _l, _p, _k, _a, _n, _s, _i)
                variance = _blrprx_variance(scale, f1, _l, _p, _k, _a, _n, _s, _i)
                if variance < 0:
                    return np.float64(np.nan)
                moment_3rd = _blrprx_moment_3rd(
                    scale, f1, f2, _l, _p, _k, _a, _n, _s, _i
                )

                for stat_idx, stat in enumerate(stats_types):
                    if weight_np[t, stat_idx] == 0:
                        continue
                    predict = 0
                    if stat == 0:  # Stat_Props.MEAN.value
                        predict = mean
                    elif stat == 1:  # Stat_Props.CVAR.value
                        predict = np.sqrt(variance) / mean
                    elif stat == 2:  # Stat_Props.SKEWNESS.value
                        predict = moment_3rd / np.power(variance, 1.5)
                    elif stat == 3:  # Stat_Props.AR1.value
                        predict = (
                            _blrprx_covariance(
                                scale, f1, 1.0, _l, _p, _k, _a, _n, _s, _i
                            )
                            / variance
                        )
                    elif stat == 4:  # Stat_Props.AR2.value
                        predict = (
                            _blrprx_covariance(
                                scale, f1, 2.0, _l, _p, _k, _a, _n, _s, _i
                            )
                            / variance
                        )
                    elif stat == 5:  # Stat_Props.AR3.value
                        predict = (
                            _blrprx_covariance(
                                scale, f1, 3.0, _l, _p, _k, _a, _n, _s, _i
                            )
                            / variance
                        )
                    elif stat == 6:  # Stat_Props.pDRY.value
                        predict = 0
                    elif stat == 7:  # Stat_Props.MSIT.value
                        predict = 0
                    elif stat == 8:  # Stat_Props.MSD.value
                        predict = 0
                    elif stat == 9:  # Stat_Props.MCIT.value
                        predict = 0
                    elif stat == 10:  # Stat_Props.MCD.value
                        predict = 0
                    elif stat == 11:  # Stat_Props.MCS.value
                        predict = 0
                    elif stat == 12:  # Stat_Props.MPC.value
                        predict = 0
                    elif stat == 13:  # Stat_Props.VAR.value
                        predict = variance
                    elif stat == 14:  # Stat_Props.AC1.value
                        predict = _blrprx_covariance(
                            scale, f1, 1.0, _l, _p, _k, _a, _n, _s, _i
                        )
                    elif stat == 15:  # Stat_Props.AC2.value
                        predict = _blrprx_covariance(
                            scale, f1, 2.0, _l, _p, _k, _a, _n, _s, _i
                        )
                    elif stat == 16:  # Stat_Props.AC3.value
                        predict = _blrprx_covariance(
                            scale, f1, 3.0, _l, _p, _k, _a, _n, _s, _i
                        )
                    score += (
                        np.power(predict - stats_np[t, stat_idx], 2)
                        * weight_np[t, stat_idx]
                    )
            # return np.nansum(np.power(predict - stats_np, 2) * weight_np * mask_np)
            return score

        return evaluation_func

    @staticmethod
    def default_stats(timescale: List[float] = [1, 3, 6, 24]) -> pd.DataFrame:
        # Create a pandas dataframe with Stat_Props as columns one row with timescales=1,3,6,24
        stats_df = pd.DataFrame(
            columns=[prop for prop in StatMetrics], index=timescale, dtype=np.float64
        )
        stats_df.index.name = "timescales"

        return stats_df

    @staticmethod
    def default_weight(timescale: List[float] = [1, 3, 6, 24]) -> pd.DataFrame:
        # Create a pandas dataframe with Stat_Props as columns one row with timescales=1,3,6,24
        weight_df = pd.DataFrame(
            columns=[prop for prop in StatMetrics], index=timescale, dtype=np.float64
        )
        weight_df.index.name = "timescales"
        return weight_df

    @staticmethod
    def default_mask(timescale: List[float] = [1, 3, 6, 24]) -> pd.DataFrame:
        # Create a numpy array with Stat_Props as columns one row with timescales=1,3,6,24, with default value of False
        mask_np = np.zeros((len(timescale), len(StatMetrics)), dtype=np.int_)
        # Create a pandas dataframe with Stat_Props as columns one row with timescales=1,3,6,24, with default value of False
        mask_df = pd.DataFrame(
            data=mask_np,
            columns=[prop for prop in StatMetrics],
            index=timescale,
            dtype=np.int_,
        )
        mask_df.index.name = "timescales"
        return mask_df


@nb.njit()
def _blrprx_kernel(k: float, u: float, nu: float, alpha: float) -> float:
    """
    k: lag
    u: timescale
    """
    # Modelling rainfall with a Bartlett–Lewis process: new developments(2020) Formula (5)

    ## TODO: Check if this is still required.
    # if alpha <= 4.0 and np.modf(alpha)[0] == 0.0:
    #    alpha += 1.0e-8

    ## TODO: Check if this is still required.
    # if alpha - k >= 171.0 or alpha >= 171.0:
    #    return np.inf
    return (
        np.power(nu / (nu + u), alpha)
        * np.power(nu + u, k)
        * math.gamma(alpha - k)
        / math.gamma(alpha)
    )


@nb.njit()
def _blrprx_mean(
    timescale: float,
    lambda_: float,
    phi: float,
    kappa: float,
    alpha: float,
    nu: float,
    sigmax_mux: float,
    iota: float,
):
    mu_c = 1.0 + kappa / phi
    # TODO: Check the formula. Why do we have a kernel here?
    return timescale * lambda_ * iota * mu_c * _blrprx_kernel(0, 0, nu, alpha)


@nb.njit()
def _blrprx_variance(
    timescale: float,
    f1: float,
    lambda_: float,
    phi: float,
    kappa: float,
    alpha: float,
    nu: float,
    sigmax_mux: float,
    iota: float,
):
    mu_c = 1.0 + kappa / phi

    # TODO: The formula is not the same as C/C++ code. Check this.
    # The formula is the same as Kaczmarska et al. (2014) A.2
    var_part1 = timescale * (f1 + kappa / phi)
    var_part2 = (kappa * (1 - phi**3) / (phi**4 - phi**2) - f1) * _blrprx_kernel(
        1.0, 0, nu, alpha
    )
    var_part3 = (kappa / (phi**4 - phi**2)) * _blrprx_kernel(
        1.0, phi * timescale, nu, alpha
    )
    var_part4 = (f1 + kappa * phi / (phi**2 - 1)) * _blrprx_kernel(
        1.0, timescale, nu, alpha
    )

    return (
        2 * lambda_ * mu_c * iota**2 * (var_part1 + var_part2 - var_part3 + var_part4)
    )


@nb.njit()
def _blrprx_covariance(
    timescale: float,
    f1: float,
    lag: float,
    lambda_: float,
    phi: float,
    kappa: float,
    alpha: float,
    nu: float,
    sigmax_mux: float,
    iota: float,
):
    mu_c = 1.0 + kappa / phi

    # TODO: In the original code there's a variable k = 3.0 - 2.0 * c.
    # Is it the same as # Kaczmarska et al. (2014) A.10?
    # The lag between covariance? Seems it doesn't.
    # Because in Kaczmarska et al. (2014) (1) There is also a k. That is not lag right?

    cov_part1 = (f1 + (kappa * phi) / (phi**2 - 1)) * (
        _blrprx_kernel(1.0, (lag - 1.0) * timescale, nu, alpha)
        - 2 * _blrprx_kernel(1.0, lag * timescale, nu, alpha)
        + _blrprx_kernel(1.0, (lag + 1.0) * timescale, nu, alpha)
    )

    cov_part2 = (kappa / (phi**4 - phi**2)) * (
        _blrprx_kernel(1.0, phi * (lag - 1.0) * timescale, nu, alpha)
        - 2 * _blrprx_kernel(1.0, phi * lag * timescale, nu, alpha)
        + _blrprx_kernel(1.0, phi * (lag + 1.0) * timescale, nu, alpha)
    )
    return lambda_ * mu_c * iota**2 * (cov_part1 - cov_part2)


@nb.njit()
def _blrprx_moment_3rd(
    timescale: float,
    f1: float,
    f2: float,
    lambda_: float,
    phi: float,
    kappa: float,
    alpha: float,
    nu: float,
    sigmax_mux: float,
    iota: float,
):
    mu_c = 1.0 + kappa / phi

    phi2, phi3, phi4, phi5, phi6, phi7, phi8, phi9 = np.power(phi, np.arange(2, 10))

    kappa2 = kappa**2

    m3_part0 = (1 + 2 * phi + phi2) * (phi4 - 2 * phi3 - 3 * phi2 + 8 * phi - 4) * phi3
    m3_part1 = _blrprx_kernel(1.0, timescale, nu, alpha) * (
        12 * phi7 * kappa2
        - 24 * f1 * phi2 * kappa
        - 18 * phi4 * kappa2
        + 24 * f1 * phi3 * kappa
        - 132 * f1 * phi6 * kappa
        + 150 * f1 * phi4 * kappa
        - 42 * phi5 * kappa2
        - 6 * f1 * phi5 * kappa
        + 108 * phi5 * f2
        - 72 * phi7 * f2
        - 48 * phi3 * f2
        + 24 * f1 * phi8 * kappa
        + 12 * phi3 * kappa2
        + 12 * phi9 * f2
    )

    m3_part2 = _blrprx_kernel(0, timescale, nu, alpha) * (
        24 * f1 * phi4 * timescale * kappa
        + 6 * f2 * phi9 * timescale
        - 30 * f1 * phi6 * timescale * kappa
        + 6 * f1 * phi8 * timescale * kappa
        + 54 * f2 * phi5 * timescale
        - 24 * f2 * phi3 * timescale
        - 36 * f2 * phi7 * timescale
    )

    m3_part3 = _blrprx_kernel(1.0, phi * timescale, nu, alpha) * (
        -48 * kappa2
        + 6 * f1 * phi4 * kappa
        - 48 * f1 * phi * kappa
        + 6 * phi5 * kappa2
        - 24 * f1 * phi2 * kappa
        + 36 * f1 * phi3 * kappa
        - 6 * f1 * phi5 * kappa
        + 84 * phi2 * kappa2
        + 12 * phi3 * kappa2
        - 18 * phi4 * kappa2
    )

    m3_part4 = _blrprx_kernel(0, phi * timescale, nu, alpha) * (
        -24 * phi * timescale * kappa2
        + 30 * phi3 * timescale * kappa2
        - 6 * phi5 * timescale * kappa2
    )

    m3_part5 = _blrprx_kernel(1.0, 0, nu, alpha) * (
        +72 * f2 * phi7
        + 48 * f1 * phi * kappa
        + 24 * f1 * phi2 * kappa
        - 36 * f1 * phi3 * kappa
        - 84 * phi2 * kappa2
        + 6 * f1 * phi5 * kappa
        + 117 * f1 * phi6 * kappa
        + 39 * phi5 * kappa2
        - 12 * f2 * phi9
        - 138 * f1 * phi4 * kappa
        + 48 * kappa2
        - 9 * phi7 * kappa2
        + 48 * f2 * phi3
        + 18 * phi4 * kappa2
        - 21 * f1 * phi8 * kappa
        - 12 * phi3 * kappa2
        - 108 * f2 * phi5
    )

    m3_part6 = timescale * (
        -24 * phi * kappa2
        - 72 * f1 * phi6 * kappa
        - 36 * phi5 * kappa2
        + 54 * phi3 * kappa2
        + 6 * phi7 * kappa2
        + 54 * f2 * phi5
        - 36 * f2 * phi7
        - 24 * f2 * phi3
        - 48 * f1 * phi2 * kappa
        + 12 * f1 * phi8 * kappa
        + 6 * f2 * phi9
        + 108 * f1 * phi4 * kappa
    )

    m3_part7 = _blrprx_kernel(1.0, 2 * timescale, nu, alpha) * (
        -12 * f1 * phi4 * kappa
        - 3 * f1 * phi8 * kappa
        + 15 * f1 * phi6 * kappa
        - 3 * phi7 * kappa2
        + 3 * phi5 * kappa2
    )

    m3_part8 = _blrprx_kernel(1.0, (1 + phi) * timescale, nu, alpha) * (
        -24 * f1 * phi3 * kappa
        - 6 * f1 * phi4 * kappa
        + 6 * f1 * phi5 * kappa
        + 24 * f1 * phi2 * kappa
        + 18 * phi4 * kappa2
        - 12 * phi3 * kappa2
        - 6 * phi5 * kappa2
    )

    return (lambda_ * mu_c * iota**3 / m3_part0) * (
        m3_part1
        + m3_part2
        + m3_part3
        + m3_part4
        + m3_part5
        + m3_part6
        + m3_part7
        + m3_part8
    )


@nb.njit()
def _blrprx_sample(
    duration_hr: float,
    intensity_sampler: Callable[
        [np.random.Generator, float, float, int], npt.NDArray[np.float64]
    ],
    rng: np.random.Generator,
    lambda_: float,
    phi: float,
    kappa: float,
    alpha: float,
    nu: float,
    sigmax_mux: float,
    iota: float,
) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.int64], npt.NDArray[np.float64]]:
    # Storm number sampling
    n_storm = rng.poisson(lambda_ * duration_hr)

    total_cells: int = 0
    eta = np.empty(n_storm, dtype=np.float64)
    mux = np.empty(n_storm, dtype=np.float64)
    storm_starts = np.empty(n_storm, dtype=np.float64)
    storm_durations = np.empty(n_storm, dtype=np.float64)
    storm_info = np.empty((n_storm, 6), dtype=np.float64)
    n_cells_per_storm = np.empty(n_storm, dtype=np.int64)
    for i in range(n_storm):
        eta[i] = rng.gamma(alpha, 1 / nu)
        gamma = phi * eta[i]
        beta = kappa * eta[i]
        mux[i] = iota * eta[i]  # mux = iota * eta
        storm_starts[i] = rng.uniform(
            0, duration_hr
        )  # storm_starts = rng.uniform(0, duration_hr, n_storm)
        storm_durations[i] = rng.exponential(
            1 / gamma
        )  # storm_durations = rng.exponential(1 / gamma, n_storm)
        n_cells_per_storm[i] = 1 + rng.poisson(
            beta * storm_durations[i]
        )  # n_cells_per_storm = 1 + rng.poisson(beta * storm_durations, size=n_storm)
        total_cells += n_cells_per_storm[i]
        storm_info[i] = (
            storm_starts[i],
            storm_durations[i],
            eta[i],
            mux[i],
            gamma,
            beta,
        )

    # Pre-allocate arrays
    cell_starts = np.empty(total_cells)  # (total_cells, )
    cell_durations = np.empty(total_cells)  # (total_cells, )
    cell_intensities = np.empty(total_cells)  # (total_cells, )

    cells_start_idx = 0
    for i, (s, d) in enumerate(zip(storm_starts, storm_durations)):
        cell_starts[cells_start_idx] = (
            s  # First cell starts at the same time as the storm
        )
        cell_starts[cells_start_idx + 1 : cells_start_idx + n_cells_per_storm[i]] = (
            rng.uniform(s, s + d, n_cells_per_storm[i] - 1)
        )

        cell_durations[cells_start_idx : cells_start_idx + n_cells_per_storm[i]] = (
            rng.exponential(scale=1 / eta[i], size=n_cells_per_storm[i])
        )

        cell_intensities[cells_start_idx : cells_start_idx + n_cells_per_storm[i]] = (
            intensity_sampler(rng, mux[i], sigmax_mux, n_cells_per_storm[i])
        )
        cells_start_idx += n_cells_per_storm[i]

    cell_ends = cell_starts + cell_durations  # (total_cells, )

    # Flatten cell_starts, cell_ends, cell_intensities and stack them together
    cell_arr = np.stack(
        (cell_starts, cell_ends, cell_intensities), axis=-1
    )  # (total_cells, 3)

    return cell_arr, n_cells_per_storm, storm_info



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from typing import Optional, Union

import numpy as np
import numpy.typing as npt
import pandas as pd  # type: ignore

from pybl.models import StatMetrics
from pybl.timeseries import IndexedSnapshot


def get_month_intervals(
    start: Optional[Union[int, datetime]] = None,
    end: Optional[Union[int, datetime]] = None,
) -> npt.NDArray[np.float64]:
    """
    ### Note that the returned unix time is `UTC`.
    Generate a 2D matrix of (nYear, 12, 2) of unix time interval of each month with its start and exclusive end unix time.

    Parameters
    ----------
    start : int, datetime, None
        Start Year, default to be 1900
    end : int, datetime, None
        End Year, default to be 2100

    Returns
    -------
    np.ndarray
        2D matrix of (nYear, 12, 2) of unix time interval of each month with its start and exclusive end unix time.
    """
    if isinstance(start, int):
        start = datetime(start, 1, 1, 0, 0, 0)
    if isinstance(end, int):
        end = datetime(end, 1, 1, 0, 0, 0)

    if start is None:
        start = datetime(1900, 1, 1, 0, 0, 0)
    if end is None:
        end = datetime(2100, 1, 1, 0, 0, 0)
    day = pd.date_range(start=start, end=end, freq="MS")
    # Convert to unix time
    month_srt = day.astype("int64") // 10**9
    month_end = month_srt
    # Stack them together
    month_intervals = np.stack((month_srt[:-1], month_end[1:]), axis=1)
    # Group the month_interval by month
    month_intervals = np.reshape(month_intervals, (-1, 12, 2))
    return month_intervals


def to_year_month(
    unix_time: npt.NDArray[np.float64], intensity: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    """
    Convert input 1D array of unix time and intensity to 2D matrix of year and month.
    With the shape of (nYear, 12). Each cell in the 2D matrix is a `IndexedSnapshot` object of that month.

    Parameters
    ----------
    unix_time : npt.NDArray
        1D array of unix time. Unit: second
    intensity : npt.NDArray
        1D array of intensity data. Unit: mm/h

    Returns
    -------
    npt.NDArray
        2D matrix of year and month full of `IndexedSnapshot` objects.
        With the shape of (nYear, 12).
    """
    series = IndexedSnapshot(unix_time, intensity / 3600)
    # Gest starting year and ending year
    start_year = datetime.fromtimestamp(unix_time[0], tz=timezone.utc).year
    end_year = datetime.fromtimestamp(unix_time[-1], tz=timezone.utc).year
    month_interval_each_year = get_month_intervals(start=start_year, end=end_year)
    seriesYearMonth = np.empty(
        (len(month_interval_each_year), 12), dtype=IndexedSnapshot
    )  # (year, month)
    for i, year in enumerate(month_interval_each_year):
        for j, month in enumerate(year):
            seriesYearMonth[i, j] = series[month[0] : month[1]]

    # Find the first row that doesn't have any IndexedSnapshot object
    for i, row in enumerate(seriesYearMonth):
        if np.isnan(row[0].mean()):
            break

    return seriesYearMonth


def calculate_stats_matrix(
    timeindex: npt.NDArray[np.float64],
    intensity: npt.NDArray[np.float64],
    timescale: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    This function calculate the statistics of the input timeseries and return the target and weight matrix for fitting.

    Parameters
    ----------
    timeindex : np.ndarray
        1D array of unix time. Unit: second

    intensity : np.ndarray
        1D array of intensity data. Unit: mm/s

    timescale : np.ndarray
        1D array of timescale. Unit: second


    Returns
    -------
    target : np.ndarray
        3D matrix of target. With the shape of (12 (month), nScale, nStats).

    weight : np.ndarray
        3D matrix of weight. With the shape of (12 (month), nScale, nStats).
    """
    timescale = np.array(timescale)

    series = IndexedSnapshot(
        timeindex, intensity
    )  # Unit: mm/h so divide by 3600 to get mm/s
    month_interval_each_year = get_month_intervals()
    # Segment the series timeseries into months from 1900 to 2100
    series_month_each = np.empty(
        (12, len(month_interval_each_year), len(timescale)), dtype=IndexedSnapshot
    )  # (month, year, scale)
    for i, year in enumerate(month_interval_each_year):
        for j, month in enumerate(year):
            for k, scale in enumerate(timescale):
                series_month_each[j, i, k] = series[month[0] : month[1]].rescale(scale)

    # IndexedSnapshot that stores the total of each month
    series_month_total = np.empty(
        (12, len(timescale)), dtype=IndexedSnapshot
    )  # (month, scale)
    for i in range(12):
        for j in range(len(series_month_each[0])):
            for k, scale in enumerate(timescale):
                if j == 0:
                    series_month_total[i, k] = IndexedSnapshot()
                series_month_total[i, k].add(
                    series_month_each[i, j, k], sequential=True
                )

    stats_month = np.zeros((12, len(timescale), 5))  # (month, scale, stats)
    for month in range(12):
        for scale in range(len(timescale)):
            model = series_month_total[month, scale]
            stats_month[month, scale, :] = [
                model.mean(),
                model.coef_variation(),
                model.autocorr_coef(1),
                model.skewness(),
                model.pDry(0),
            ]

    stats_month_seperate = np.zeros(
        (12, len(month_interval_each_year), len(timescale), 5)
    )  # (month, year, scale, stats)
    for month in range(12):
        for year in range(len(month_interval_each_year)):
            for scale in range(len(timescale)):
                model = series_month_each[month, year, scale]
                stats_month_seperate[month, year, scale, :] = [
                    model.mean(),
                    model.coef_variation(),
                    model.autocorr_coef(1),
                    model.skewness(),
                    model.pDry(0),
                ]

    stats_weight = 1 / np.nanvar(stats_month_seperate, axis=1)  # (month, scale, stats)

    return stats_month, stats_weight


def classic_statistics(
    timeseries: pd.Series[float], timescale: list[timedelta], intensity_time_interval: timedelta
) -> tuple[list[pd.DataFrame], list[pd.DataFrame]]:
    """
    This function calculate the statistics of the input timeseries and return the target and weight matrix for fitting.

    Parameters
    ----------
    timeseries : pd.Series
        The input timeseries data. The index of the series is datetime and the value is the intensity. The unit of the intensity should be mm/h.

    timescale : list[timedelta]
        The list of timescale that we want to calculate the statistics.

    Returns
    -------
    target : pd.DataFrame
        The target matrix for fitting. The columns are the statistics properties and the rows are the month and timescale.

    weight : pd.DataFrame
        The weight matrix for fitting. The columns are the statistics properties and the rows are the month and timescale.
    """

    scale_second = np.array([ts.total_seconds() for ts in timescale], dtype=np.float64)
    scale_hr = scale_second / 3600

    # Copy the timeseries to avoid changing the original one
    timeseries = timeseries.copy()

    # Check the index type of the timeseries. It can be either DatetimeIndex or PeriodIndex
    if isinstance(timeseries.index, pd.DatetimeIndex):
        pass
    elif isinstance(timeseries.index, pd.PeriodIndex):
        timeseries.index = timeseries.index.to_timestamp()
    else:
        raise ValueError(f"The index of the timeseries should be either DatetimeIndex or PeriodIndex. Got {type(timeseries.index)}")

    # Check the type of the intensity data. It should be a number that can be converted to float
    if not pd.api.types.is_numeric_dtype(timeseries):
        raise ValueError(f"The intensity data should be numeric. Got {timeseries.dtype}")


    unix_time = timeseries.index.to_numpy().astype(np.float64) // 10**9
    intensity = timeseries.to_numpy()
    # Replace all the negative intensity to np.nan
    intensity[intensity < 0] = np.nan
    intensity = intensity / intensity_time_interval.total_seconds()

    target, weight = calculate_stats_matrix(unix_time, intensity, scale_second)

    target_df_list = []
    weight_df_list = []
    for month_idx in range(12):
        # Remove pDry from the target and weight
        target_df = pd.DataFrame(
            target[month_idx, :, :-1], columns=[StatMetrics.MEAN, StatMetrics.CVAR, StatMetrics.AR1, StatMetrics.SKEWNESS], index=scale_hr
        )
        weight_df = pd.DataFrame(
            weight[month_idx, :, :-1], columns=[StatMetrics.MEAN, StatMetrics.CVAR, StatMetrics.AR1, StatMetrics.SKEWNESS], index=scale_hr
        )
        target_df.index.name = "timescale_hr"
        weight_df.index.name = "timescale_hr"

        target_df_list.append(target_df)
        weight_df_list.append(weight_df)



    return target_df_list, weight_df_list



from typing import Optional, Union, overload

import numpy as np
import numpy.typing as npt
import pandas as pd


@overload
def generate(intensity: pd.Series[float], sample_size: int) -> pd.Series[float]: ...


@overload
def generate(
    intensity: npt.NDArray[np.float64],
    sample_size: int,
    epoch_time: Optional[npt.NDArray[np.float64]],
) -> npt.NDArray[np.float64]: ...


def generate(
    intensity: Union[pd.Series[float], npt.NDArray[np.float64]],
    sample_size: int,
    *args,
    **kwargs
) -> Union[pd.Series[float], npt.NDArray[np.float64]]:
    return np.empty(0)



from __future__ import annotations

import math
from typing import Protocol, runtime_checkable

import numba as nb  # type: ignore
import numpy as np
import numpy.typing as npt


@runtime_checkable
class IConstantRCI(Protocol):
    @staticmethod
    def get_f1(sigmax_mux: float) -> float:
        ...

    @staticmethod
    def get_f2(sigmax_mux: float) -> float:
        ...

    @staticmethod
    def sample_intensity(
        rng: np.random.Generator,
        mux: float,
        sigmax_mux: float,
        n_cells: int,
    ) -> npt.NDArray[np.float64]:
        ...


class GammaRCIModel(IConstantRCI):
    @staticmethod
    @nb.njit
    def get_f1(sigmax_mux: float) -> float:
        return (sigmax_mux + 1.0) / sigmax_mux

    @staticmethod
    @nb.njit
    def get_f2(sigmax_mux: float) -> float:
        x2 = sigmax_mux**2
        return (x2 + 3.0 * sigmax_mux + 2.0) / x2

    @staticmethod
    @nb.njit
    def sample_intensity(
        rng: np.random.Generator, mux: float, sigmax_mux: float, n_cells: int = 1
    ) -> npt.NDArray[np.float64]:
        return rng.gamma(sigmax_mux, mux / sigmax_mux, size=n_cells)


class ExponentialRCIModel(IConstantRCI):
    @staticmethod
    @nb.njit
    def get_f1(sigmax_mux: float) -> float:
        return 2.0

    @staticmethod
    @nb.njit
    def get_f2(sigmax_mux: float) -> float:
        return 6.0

    @staticmethod
    @nb.njit
    def sample_intensity(
        rng: np.random.Generator, mux: float, sigmax_mux: float, n_cells: int = 1
    ) -> npt.NDArray[np.float64]:
        I_shape = 1.0 / sigmax_mux**2
        I_scale = sigmax_mux**2 * mux
        return rng.gamma(I_shape, I_scale, size=n_cells)


class WeibullRCIModel(IConstantRCI):
    @staticmethod
    @nb.njit
    def get_f1(sigmax_mux: float) -> float:
        ex2: float = math.gamma(1.0 + 2.0 / sigmax_mux)
        ex: float = math.gamma(1.0 + 1.0 / sigmax_mux)
        return ex2 / ex**2

    @staticmethod
    @nb.njit
    def get_f2(sigmax_mux: float) -> float:
        ex3: float = math.gamma(1.0 + 3.0 / sigmax_mux)
        ex: float = math.gamma(1.0 + 1.0 / sigmax_mux)
        return ex3 / ex**3

    @staticmethod
    @nb.njit
    def sample_intensity(
        rng: np.random.Generator, mux: float, sigmax_mux: float, n_cells: int = 1
    ) -> npt.NDArray[np.float64]:
        I_scale = mux / math.gamma(1.0 + 1.0 / sigmax_mux)
        # The weibull distribution in numpy differs from the one in c++
        # See Numpy: https://numpy.org/doc/stable/reference/random/generated/numpy.random.Generator.weibull.html
        # See GSL: https://www.gnu.org/software/gsl/doc/html/randist.html#c.gsl_ran_weibull
        return rng.weibull(sigmax_mux, size=n_cells) * I_scale


class ParetoRCIModel(IConstantRCI):
    @staticmethod
    @nb.njit
    def get_f1(sigmax_mux: float) -> float:
        return (sigmax_mux - 1.0) ** 2 / sigmax_mux / (sigmax_mux - 2.0)

    @staticmethod
    @nb.njit
    def get_f2(sigmax_mux: float) -> float:
        return (sigmax_mux - 1.0) ** 3 / sigmax_mux**2 / (sigmax_mux - 3.0)

    @staticmethod
    @nb.njit
    def sample_intensity(
        rng: np.random.Generator, mux: float, sigmax_mux: float, n_cells: int = 1
    ) -> npt.NDArray[np.float64]:
        I_scale = mux * (sigmax_mux - 1.0) / sigmax_mux
        # The pareto distribution in numpy differs from the one in c++
        # See Numpy: https://numpy.org/doc/stable/reference/random/generated/numpy.random.Generator.pareto.html
        # See GSL: https://www.gnu.org/software/gsl/doc/html/randist.html#c.gsl_ran_pareto
        return rng.pareto(sigmax_mux, size=n_cells) * I_scale + I_scale


class GPRCIModel(IConstantRCI):
    @staticmethod
    @nb.njit
    def get_f1(sigmax_mux: float) -> float:
        return 2.0 * (1.0 - sigmax_mux) / (1.0 * 2.0 * sigmax_mux)

    @staticmethod
    @nb.njit
    def get_f2(sigmax_mux: float) -> float:
        return (
            6.0
            * (sigmax_mux - 1.0) ** 2
            / (1.0 - 3.0 * sigmax_mux)
            / (1.0 - 2.0 * sigmax_mux)
        )

    @staticmethod
    @nb.njit
    def sample_intensity(
        rng: np.random.Generator, mux: float, sigmax_mux: float, n_cells: int = 1
    ) -> npt.NDArray[np.float64]:
        I_scale = mux * (1.0 - sigmax_mux)
        I_location = 0.0

        u = 1.0 - rng.uniform()

        if sigmax_mux == 0.0:
            intensity = I_location - I_scale * np.log(u)
        else:
            intensity = (
                I_location
                + I_scale * (np.power(u, -1.0 * sigmax_mux) - 1.0) / sigmax_mux
            )

        return intensity


class ConstantRCI:
    GammaRCIModel = GammaRCIModel
    ExponentialRCIModel = ExponentialRCIModel
    WeibullRCIModel = WeibullRCIModel
    ParetoRCIModel = ParetoRCIModel
    GPRCIModel = GPRCIModel



from __future__ import annotations

from enum import Enum
from typing import Optional, Type, Union

import numpy as np

from pybl.raincell import ExponentialRCIModel, IConstantRCI

BaseBLRP_RCIModel = Optional[Union[IConstantRCI, Type[IConstantRCI]]]


class StatMetrics(Enum):
    MEAN = 0  # Mean
    CVAR = 1  # Coefficient of variation
    SKEWNESS = 2  # Skewness
    AR1 = 3  # Autocorrelation coefficient lag 1
    AR2 = 4  # Autocorrelation coefficient lag 2
    AR3 = 5  # Autocorrelation coefficient lag 3
    pDRY = 6  # Probability of dry
    MSIT = 7
    MSD = 8
    MCIT = 9
    MCD = 10
    MCS = 11
    MPC = 12
    VAR = 13
    AC1 = 14  # Autocorrelation lag 1 (Just covariance with different lags)
    AC2 = 15  # Autocorrelation lag 2 (Just covariance with different lags)
    AC3 = 16  # Autocorrelation lag 3 (Just covariance with different lags)


class BaseBLRP():
    rci_model: IConstantRCI
    rng: np.random.Generator

    def __init__(
        self,
        rng: Optional[np.random.Generator] = None,
        rci_model: BaseBLRP_RCIModel = None,
    ) -> None:
        if rng is None:
            self.rng = np.random.default_rng()
        elif isinstance(rng, np.random.Generator):
            self.rng = rng
        else:
            self.rng = np.random.default_rng(rng)  # type: ignore

        if rci_model is None:
            self.rci_model = ExponentialRCIModel()
        elif isinstance(rci_model, IConstantRCI):
            self.rci_model = rci_model
        else:
            raise TypeError("rci_model must be a implementation of IConstantRCI")
